<?php
class ControllerFeedGoDataFeed extends Controller {
  // Holds
  private $dom;
  private $root;
  private $fields;
  private $fields_list = array('Product_ID', 'Parent_ID', 'Sku', 'Name', 'Description', 'Price', 'MerchantCategory', 'URL', 'ImageURL', 'Manufacturer', 'ManufacturerPartNumber', 'Brand', 'Keywords', 'Quantity', 'Weight', 'UPC', 'SalePrice', 'GDFVariations');
  private $catalog;
  private $products;

  // Repeated fields thresholds. Set to 0 to disable.
  private $MAX_ADDITIONAL_IMAGES = 4;
  private $MAX_DISCOUNTPRICEFIELDS = 0;
  private $MAX_DISCOUNTQUANTITYFIELDS = 0;
  /**
   * ControllerFeedGoDataFeed::index()
   * action for controller
   *
   * @return void
   */
  public function index() {
    if ($this->config->get('godatafeed_status')) {
      // Load all models
      $this->load->model('tool/image');
      $this->load->model('catalog/product');
      $this->load->model('catalog/category');
      if ($this->isModelAvailable('tool/seo_url')) {
        $this->load->model('tool/seo_url');
      }

      // Create new dom docuemnt and root and assign them to the private vars
      $this->dom = new DOMDocument('1.0', 'UTF-8');
      $this->root = $this->_c('GoDataFeed');
      $this->root = $this->dom->appendChild($this->root);

      // Set up pagination
      $paging_start = isset($this->request->get['start'])?$this->request->get['start']:0;
      $paging_count = isset($this->request->get['count'])?$this->request->get['count']:0;
      $is_paging = $paging_start > 0 && $paging_count > 0;
      $paging_total = 0;

      // Setting parameters for retrieving list of products
      if ($is_paging) {
        $data = array(
            'start' => $paging_start-1,
            'limit' => $paging_count
        );
      } else {
        $data = array();
      }
      // Load products
      $products = $this->model_catalog_product->getProducts($data);
      $paging_total = $product_total = $this->model_catalog_product->getTotalProducts($data);

      // Enforce pagination for old OpenCart versions that don't support $data in getProducts()
      if ($is_paging && count($products) > $paging_count) {
        $products = array_slice($products, $paging_start - 1, $paging_count);
      }

      if ($is_paging) {
        $this->paging = $this->_c('Paging');
        $this->paging->appendChild($this->_c('Start', $paging_start));
        $this->paging->appendChild($this->_c('Count', $paging_count));
        $this->paging->appendChild($this->_c('Total', $paging_total));
        $this->paging = $this->root->appendChild($this->paging);
      }

      // Set up fields
      $this->fields = $this->_c('Fields');
      $this->fields = $this->root->appendChild($this->fields);
      $this->addFields($this->fields, $this->fields_list);

      // Set up product nodes
      $this->catalog = $this->_c('Products');
      $this->catalog = $this->root->appendChild($this->catalog);

      foreach($products as  $p) {
        $product_options = $this->model_catalog_product->getProductOptions($p['product_id']);
        if (count($product_options)) {
          $item = $this->getProductNode($p, TRUE);
          $product_options_combos = $this->getOptions($p['product_id']);
          foreach ($product_options_combos as $options_combo) {
            $tmpItem = $item->cloneNode(TRUE);

            // appending the combo-name to the ID
            $tmpItemID = $tmpItem->getElementsByTagName('Product_ID')->item(0);
            $itemOptID = $this->_c('Product_ID', $p['product_id'] . "_" . $options_combo['combo_name'], false);
            $tmpItem->replaceChild($itemOptID, $tmpItemID);

            // updating price to reflect options-combo difference
            $tmpItemPrice = $tmpItem->getElementsByTagName('Price')->item(0);
            $itemOptPrice = $this->getPriceNode($p, $options_combo['price_diff']);
            $tmpItem->replaceChild($itemOptPrice, $tmpItemPrice);

            // updating sale-price to reflect options-combo difference
            $tmpItemSalePrice = $tmpItem->getElementsByTagName('SalePrice')->item(0);
            $itemOptSalePrice = $this->getSalePriceNode($p, $options_combo['price_diff']);
            $tmpItem->replaceChild($itemOptSalePrice, $tmpItemSalePrice);

            // creating custom elements for product-options
            foreach ($options_combo['combo_options'] as $combo_opt_name => $combo_opt_val) {
              $tmpItem->appendChild($this->_c($combo_opt_name, $combo_opt_val));

              // add the custom option's name to the Fields list
              $this->addCustomField($this->fields, $combo_opt_name);
            }


            $this->catalog->appendChild($tmpItem);
          }
        } else {
          $this->catalog->appendChild($this->getProductNode($p));
        }
      }

      $this->dom->formatOutput = true;
      $this->response->addHeader('Content-Type: application/xml');
      $this->response->setOutput($this->dom->saveXML());
    } else {
      die('This module needs to be enabled in the admin interface');
    }
  }


  /**
   * ControllerFeedGoDataFeed::getCategories()
   * Adds the list of categories to the GDF feed recursively
   *
   * @param integer $parent_id
   * @param string $current_path
   * @return void
   */
  protected function getCategories($parent_id, $current_path = '') {
    // Load categories
    $results = $this->model_catalog_category->getCategories($parent_id);

    // Loop through each category, increasing the path as the structure gets deeper
    foreach ($results as $result) {
      if (!$current_path) {
        $new_path = $result['category_id'];
      } else {
        $new_path = $current_path . '_' . $result['category_id'];
      }
      // add category url
      $this->addUrl(HTTP_SERVER . 'index.php?route=product/category&path=' . $new_path, '0.7');
      // Recursion to load new category level
      $this->getCategories($result['category_id'], $new_path);
    }
  }

  protected function getOptions($product_id) {
    $result = array();
    $product_options = $this->model_catalog_product->getProductOptions($product_id);
    if (count($product_options)) {
      $option_attributes = array();
      foreach ($product_options as $prod_opt) {
        $option_values = array();
        $opt_name = $this->sanitizeString($prod_opt['name']);
        foreach ($prod_opt['product_option_value'] as $opt_val) {
          $price_diff = (float)$opt_val['price'];
          $price_diff *= (($opt_val['prefix'] == "-") ? -1 : 1);
          $option_values[] = array("name" => $this->sanitizeString($opt_val['name']), "price_diff" => $price_diff, "opt_name" => $opt_name);
        }
        $option_attributes[] = $option_values;
      }

      $result = $this->getArrayCombos($option_attributes);
    }

    return $result;
  }

  protected function sanitizeString($input) {
    $result = ucwords(trim($input));
    $result = preg_replace("/[^a-z0-9]/i", "-", $result);
    return $result;
  }

  /*** NOTE: This has been customized SPECIFICALLY for an array of product-option attributes! ***/
  /** Source for original: https://gist.github.com/fabiocicerchia/4556892 **/
  protected function getArrayCombos(array $data, array &$all = array(), array $group = array(), $value = null, $i = 0) {
    $keys = array_keys($data);
    if (isset($value) === true) {
      array_push($group, $value);
    }

    if ($i >= count($data)) {
      $tmpArr = array();
      $total_price_diff = (float)0;
      $optionsArr = array();
      foreach ($group as $g) {
        $tmpArr[] = $g['name'];
        $total_price_diff += $g['price_diff'];
        $optionsArr[$g['opt_name']] = $g['name'];
      }
      array_push($all, array("combo_name" => implode("_", $tmpArr), "combo_price_diff" => $total_price_diff, "combo_options" => $optionsArr));
    } else {
      $currentKey = $keys[$i];
      $currentElement = $data[$currentKey];
      foreach ($currentElement as $val) {
        $this->getArrayCombos($data, $all, $group, $val, $i + 1);
      }
    }

    return $all;
  }

  protected function _c($name, $value = NULL, $already_raw = true) {
    if($value === NULL) return $this->dom->createElement($name);
    if(!$already_raw) {
      $value = strip_tags(html_entity_decode($value, ENT_QUOTES, 'UTF-8'));
      $value = str_replace('&nbsp;', ' ', $value);
      $value = str_replace('"', '&#34;', $value );
      $value = str_replace("'", '&#39;', $value );
      $value = str_replace('<', '&lt;', $value );
      $value = str_replace('>', '&gt;', $value );
      $value = str_replace("\n", ' ', $value );
      $value = str_replace("\r", ' ', $value );
      $value = str_replace("", '', $value ); // This special character rarely appears, but just in case...
      $value = preg_replace('/&#?[a-z0-9]+;/i',' ',$value); // remove any html entities..
      $value = preg_replace('/\s{2,}/i', ' ', $value );
      $value = substr($value, 0, 5000 );
    }

    //return $this->dom->createElement($name, $value);
    //return $this->dom->createElement($name, htmlentities($value, ENT_QUOTES, 'UTF-8'));
    $ret = $this->dom->createElement($name);
    $ret->appendChild($this->dom->createCDATASection($value));
    return $ret;
  }

  protected function getPath($parent_id, $current_path = '') {
    $category_info = $this->model_catalog_category->getCategory($parent_id);

    if ($category_info) {
      if (!$current_path) {
        $new_path = $category_info['category_id'];
      } else {
        $new_path = $category_info['category_id'] . '_' . $current_path;
      }

      $path = $this->getPath($category_info['parent_id'], $new_path);

      if ($path) {
        return $path;
      } else {
        return $new_path;
      }
    }
  }

  protected function addFields($parent_node, &$list) {
    // Additional Images
    for ($i = 1; $i <= $this->MAX_ADDITIONAL_IMAGES; $i++) {
      $list[] = 'AdditionalImage' . $i;
    }

    // Set up Discount Quantity fields
    for ($i = 1; $i <= $this->MAX_DISCOUNTQUANTITYFIELDS; $i++) {
      $list[] = 'DiscountQuantity' . $i;
    }

    // Set up Discount Price fields
    for ($i = 1; $i <= $this->MAX_DISCOUNTPRICEFIELDS; $i++) {
      $list[] = 'DiscountPrice' . $i;
    }

    foreach ($list as $field_name) {
      $field = $this->_c('Field');
      $field->setAttribute('name', $field_name);
      $parent_node->appendChild($field);
    }
  }

  /**
   * addCustomField checks if a field exists and if not creates it
   */
  protected function addCustomField($parent_node, $field_name) {
    $xpath = new DomXpath($this->dom);
    if ($xpath->query('//Field[@name="'.$field_name.'"]')->length == 0) {
      $field = $this->_c('Field');
      $field->setAttribute('name', $field_name);
      $parent_node->appendChild($field);
    }
  }

  protected function isModelAvailable($model) {
    $file  = DIR_APPLICATION . 'model/' . $model . '.php';
    return file_exists($file);
  }

  /*
   * getProductNode = return an XML node from a product object
   */
  protected function getProductNode($p, $has_options = FALSE) {
    $item = $this->_c('Product');
    $product_images = $this->model_catalog_product->getProductImages($p['product_id']);
    $product_discounts = $this->model_catalog_product->getProductDiscounts($p['product_id']);

    foreach ($this->fields_list as $field_name) {
      switch ($field_name) {
        case 'Product_ID':
          $item->appendChild($this->_c($field_name, $p['product_id'], false));
          break;
        case 'Parent_ID':
          $item->appendChild($this->_c($field_name, $has_options ? $p['product_id'] : '', false));
          break;
        case 'Sku':
          $item->appendChild($this->_c($field_name, $p['sku'], false));
          break;
        case 'Name':
          $item->appendChild($this->_c($field_name, $p['name'], false));
          break;
        case 'Description':
          $item->appendChild($this->_c($field_name, $p['description'], false));
          break;
        case 'Price':
          $item->appendChild($this->getPriceNode($p));
          break;
        case 'SalePrice':
          $item->appendChild($this->getSalePriceNode($p));
          break;
        case 'MerchantCategory':
          $categories = $this->model_catalog_product->getCategories($p['product_id']);
          $categs = array(); // used to concatenate all categories, including paths, into one string
          foreach($categories as $c) {
            $path = $this->getPath($c['category_id']);
            if ($path) {
              $cats = array();
              foreach (explode('_', $path) as $path_id) {
                $category_info = $this->model_catalog_category->getCategory($path_id);
                if ($category_info) {
                  $cats[] = $category_info['name'];
                }
              }
              if(count($cats)) $categs[] = implode(' > ', $cats);
            }
          }
          if(count($categs)) {
            $item->appendChild($this->_c($field_name, implode(', ', $categs), false));
          }
          break;
        case 'URL':
          if ($this->model_tool_seo_url) {
            $urlField = $this->_c($field_name, $this->model_tool_seo_url->rewrite(HTTP_SERVER . 'index.php?route=product/product&product_id='.$p['product_id']), false);
          } else if ($this->url) {
            $urlField = $this->_c($field_name, $this->url->link('product/product', 'product_id=' . $p['product_id']));
          } else {
            $urlField = $this->_c($field_name, "N/A");
          }
          $item->appendChild($urlField);
          break;
        case 'ImageURL':
          $image = $this->model_tool_image->resize(($p['image'] ? $p['image'] : 'no_image.jpg'), 500, 500);
          $item->appendChild($this->_c($field_name, $image));
          break;
        case 'Manufacturer':
          $item->appendChild($this->_c($field_name, $p['manufacturer'], false));
          break;
        case 'ManufacturerPartNumber':
          $item->appendChild($this->_c($field_name, $p['model']));
          break;
        case 'Brand':
          // Using Manufacturer for Brand: OpenCart does not have a Brand field for products, but it is a required field by the GDF spec
          $item->appendChild($this->_c($field_name, $p['manufacturer'], false));
          break;
        case 'Keywords':
          // Add Keywords from the product's Tags field (limiting to the first 10 to match GDF spec)
          /*$tags = $this->model_catalog_product->getProductTagsById($p['product_id']);
          $ts = "";
          if ($tags) {
            $ts = array();
            foreach($tags as $t) {
              $ts[] = $t['tag'];
            }

            if (count($ts) > 10) array_splice($ts, 10);
            $ts = implode(',', $ts);
          }
          $item->appendChild($this->_c($field_name, $ts, false));*/
          $item->appendChild($this->_c($field_name, $p['tag'], false)); //$tags = explode(',', $p['tag']);
          break;
        case 'Quantity':
          $item->appendChild($this->_c($field_name, $p['quantity']));
          break;
        case 'Weight':
          //$item->appendChild($this->_c('Weight', $this->weight->format($p['weight'], $p['weight_class'])));
          $item->appendChild($this->_c($field_name, $p['weight']));
          break;
        case 'UPC':
          $item->appendChild($this->_c($field_name, $p['model']));
          break;
        case (strpos($field_name, 'AdditionalImage') !== false && preg_match('/\d+/', $field_name, $img_matches) > 0):
          $img_num = intval($img_matches[0]) - 1;
          if (isset($product_images[$img_num])) {
            $img = $this->model_tool_image->resize(($product_images[$img_num]['image'] ? $product_images[$img_num]['image'] : 'no_image.jpg'), 500, 500);
            $item->appendChild($this->_c($field_name, $img));
          } else {
            $item->appendChild($this->_c($field_name));
          }
          break;
        case (strpos($field_name, 'DiscountQuantity') !== false && preg_match('/\d+/', $field_name, $dquantity_matches) > 0):
          $dquantity_num = intval($dquantity_matches[0]) - 1;
          if (isset($product_discounts[$dquantity_num])) {
            $dquantity = $product_discounts[$dquantity_num]['quantity'];
            $item->appendChild($this->_c($field_name, $dquantity));
          } else {
            $item->appendChild($this->_c($field_name));
          }
          break;
        case (strpos($field_name, 'DiscountPrice') !== false && preg_match('/\d+/', $field_name, $dprice_matches) > 0):
          $dprice_num = intval($dprice_matches[0]) - 1;
          if (isset($product_discounts[$dprice_num])) {
            $dprice = $product_discounts[$dprice_num]['price'];
            $item->appendChild($this->_c($field_name, number_format(round($dprice, 2), 2, '.', '')));
          } else {
            $item->appendChild($this->_c($field_name));
          }
          break;
        default:
          // handle cases where a field was added to the list but not handled above by creating an empty node
          $item->appendChild($this->_c($field_name));
          break;
      }
    }

    return $item;
  }

  protected function getPriceNode($product, $product_option_combo_price_diff = 0) {
    $price = $this->tax->calculate($product['price'] + $product_option_combo_price_diff, $product['tax_class_id']);
    $price = number_format(round($price, 2), 2, '.', '');
    return $this->_c('Price', $price);
  }

  protected function getSalePriceNode($product, $product_option_combo_price_diff = 0) {
    //$special = $this->model_catalog_product->getProductSpecial($p['product_id']);
    $special = $product['special'];
    if ($special) {
      $special = $this->tax->calculate($special + $product_option_combo_price_diff, $product['tax_class_id']);
      $special = number_format(round($special, 2), 2, '.', '');
    } else $special = '';
    return $this->_c('SalePrice', $special);
  }
}
?>